# File svdfit/Src/eigenfit2.x
# November 20, 2006
# By Doug Mink, Harvard-Smithsonian Center for Astrophysics
 
# EIGENFIT2 fits two sets of eigenvectors to a spectrum and its residuals
# from the first fit, returning the residuals from the second fit

# Copyright(c) 2000 Smithsonian Astrophysical Observatory
# You may do anything you like with this file except remove this copyright.
# The Smithsonian Astrophysical Observatory makes no representations about
# the suitability of this software for any purpose.  It is provided "as is"
# without express or implied warranty.
 
include	<imhdr.h>
include	<imio.h>
include	<fset.h>
include <smw.h>
include "rvsao.h"
include	"xcv.h"
include "contin.h"

procedure eigenfit2 (specfile,specdir,specap,specband,nbin,rebspec,resspec1,
		    resspec2, fitspec,rebwl,ispec,nspec,imout,outpath,cfd1,cfd2)

char	specfile[ARB]	# Object spectrum file name
char	specdir[ARB]	# Directory for object spectra
int	specap		# Object aperture to read from multispec file
int	specband	# Object band to read from multispec file
int	nbin		# Number of pixels in rebinned spectrum and eigenvectors
real	rebspec[ARB]	# Rebinned spectrum (filled and returned)
real	resspec1[ARB]	# First pass spectrum residuals (filled and returned)
real	resspec2[ARB]	# Second pass spectrum residuals (filled and returned)
real	fitspec[ARB]	# Fit eigenvectors (filled and returned)
real	rebwl[ARB]	# Wavelength vector for rebinned spectra
int	ispec		# Number of this spectrum
int	nspec		# Number of spectra in output file
pointer	imout		# Output data structure, NULL on first spectrum per file
char	outpath[ARB]	# Output file name
int	cfd1,cfd2	# File descriptors for coefficient file

pointer spectrum	# Object spectrum
pointer specim		# Object image header structure

int	i
char	specpath[SZ_PATHNAME]	# Object spectrum path name
char	tempfiles[SZ_PATHNAME]	# List of eigenvector spectra
char	tempfile[SZ_PATHNAME]	# Template spectrum file name
char	tempdir[SZ_PATHNAME]	# Directory for eigenvector spectra
char	tempnums0[SZ_LINE]	# List of eigenvector numbers to use
char	fstring[SZ_LINE]	# Type of continuum subtracted
char	xlab[SZ_LINE]		# Title for wavelength plots of spectrum

pointer	specwcs			# Pointer to first data point with WCS
pointer	wlwcs			# Pointer to first wavelength with WCS
int	npspec			# Number of spectrum points with WCS

double	renorm1		# renormalize before first fit by dividing by mean/this
bool	pltres1, pltres2

double 	minwav0,maxwav0	# Starting,ending wavelengths from parameter file
double 	minwav,maxwav	# Wavelength limits for cross-correlation
double	dwav		# Wavelength per pixel
double	spwl1,spwl2	# Spectrum first and last wavelengths in angstroms
int	nfvec1		# Number of pass 1 eigenvectors actually fit
int	nfvec2		# Number of pass 2 eigenvectors actually fit
int	tband		# Multispec band for eigenvectors
int	ivec, iv
int	ier
char	keyword[16]
char	tstring[SZ_LINE]

char	title[SZ_LINE]

int	spix1, spix2
double	dindef
double	tmshft, spshft
double	chi, chisq
double	coeff1[MAXTEMPS]
double	coeff2[MAXTEMPS]
int	nread
 
bool	clgetb()
int	clgeti()
double	clgetd()
int	imaccess(), strlen(), ldir
double	wcs_p2w()

include	"rvsao.com"
include "results.com"
include "contin.com"
include "svdvec.com"

define	endfit_  90
 
begin
	c0 = 299792.5d0
	schop = FALSE
	tchop = FALSE
	tmshft = 0.d0
	dindef = INDEFD
	specsh = NULL
	pltres1  = clgetb ("res1_plot")
	pltres2  = clgetb ("res2_plot")
	renorm1 = clgetd ("renorm1")
 
# Optional wavelength limits for eigenvector fit
	minwav0 = clgetd ("st_lambda")
	if (minwav0 == 0.d0)
	    minwav0 = dindef
	maxwav0 = clgetd ("end_lambda")
	if (maxwav0 == 0.d0)
	    maxwav0 = dindef

	if (debug) {
	    call printf ("EIGENFIT2: nbin = %d, nspec = %d\n")
		call pargi (nbin)
		call pargi (nspec)
	    }

# Set the object spectrum path
	ldir = strlen (specdir)
	if (ldir > 0) {
	    if (specdir[ldir] != '/') {
		specdir[ldir+1] = '/'
		specdir[ldir+2] = EOS
		}
	    call strcpy (specdir,specpath,SZ_PATHNAME)
	    call strcat (specfile,specpath,SZ_PATHNAME)
	    }
	else
	    call strcpy (specfile,specpath,SZ_PATHNAME)

# Check for readability of object spectrum
	if (imaccess (specpath, READ_ONLY) == NO) {
	    call eprintf ("EIGENFIT2: cannot read spectrum path %s \n")
		call pargstr (specpath)
	    return
	    }

# Load the object spectrum
	call getspec (specpath, specap, specband, spectrum, specim, FALSE)
	if (specim == ERR) {
	   call eprintf ("EIGENFIT2: Error reading spectrum %s\n")
	 	call pargstr (specpath)
	    go to endfit_
	    }

	call wcs_set (specsh)
	spix1 = NP1(specsh)
	spix2 = NP2(specsh)
	spwl1 = wcs_p2w (double (spix1))
	spwl2 = wcs_p2w (double (spix2))
	spdwl = (spwl2 - spwl1) / double (spix2 - spix1)

	call strcpy (LABEL(specsh), xlab, SZ_LINE)
	if (strlen (xlab) <= 0)
	    call strcpy ("Wavelength", xlab, SZ_LINE)
	if (strlen (UNITS(specsh)) > 0) {
	    call strcat (" in ",xlab,SZ_LINE)
	    call strcat (UNITS(specsh),xlab,SZ_LINE)
	    }
	else
	    call strcat (" in Angstroms",xlab,SZ_LINE)

# Renormalize spectrum if requested
	call renormspec (Memr[spectrum], renorm1, specpix, ier)
	if (ier > 0)
	    go to endfit_

# Set pixel limits of spectrum WCS
	specwcs = spectrum + NP1(specsh) - 1
	wlwcs = SX(specsh) + NP1(specsh) - 1
	npspec = NP2(specsh) - NP1(specsh) + 1

# If plot enabled, show the object spectrum.
	if (pltspec) {
	    if (nspec > 0) {
		call sprintf (title, SZ_LINE, "%d: %s original")
		    call pargi (ispec)
		    call pargstr (specname)
		}
	    else {
		call sprintf (title, SZ_LINE, "%s original")
		    call pargstr (specname)
		}
	    call plotspec (npspec,Memr[specwcs],title,Memr[wlwcs],xlab,nsmooth)
	    }

	if (debug) {
	    call printf ("EIGENFIT2: nbin = %d -> %d title= %s\n")
		call pargi (npspec)
		call pargi (nbin)
		call pargstr (title)
	    }

# Eignenvectors to which to fit spectra
	call clgstr ("eigenfile1",tempfiles,SZ_PATHNAME)
	call clgstr ("eigendir1",tempdir,SZ_PATHNAME)
	ldir = strlen (tempdir)
	if (ldir > 0 && tempdir[ldir] != '/' && tempdir[ldir] != '$') {
	    tempdir[ldir+1] = '/'
	    tempdir[ldir+2] = EOS
	    }

# Band (object, raw, sky) of eigenvector spectrum to use
	tband = clgeti ("eigenband1")

# Multispec eigenvector numbers (use only first if multiple files)
	call clgstr ("eigenum1", tempnums0, SZ_LINE)

# Load first pass eigenvectors
	minwav = minwav0
	maxwav = maxwav0
	call getvecs (tempfiles,tempdir,tband,tempnums0,minwav,maxwav,dwav,
		      spwl1, spwl2, rebvec1, nbin, rebwl, 1, nread)
	if (nread == 0)
	    go to endfit_

# Process object spectrum
	if (debug) {
	    call printf("EIGENFIT2: Overlap %9.4f to %9.4f by %9.4f, nbin = %d \n")
		call pargd (minwav)
		call pargd (maxwav)
		call pargd (dwav)
		call pargi (nbin)
	    call printf("EIGENFIT2: Spectrum %9.4f to %9.4f by %9.4f = %d %d\n")
		call pargd (spwl1)
		call pargd (spwl2)
		call pargd (spdwl)
		call pargi (SN(specsh))
		call pargi (DC(specsh))
	    }

	# Rebin spectrum to overlap these eigenvectors
	spshft = 0.d0
	if (spwl1 == minwav && spdwl == dwav && specpix == nbin) {
	    if (debug) {
		call printf ("EIGENFIT2: copying %d-pixel spectrum\n")
		    call pargi (nbin)
		call flush (STDOUT)
		}
	    call amovr (Memr[spectrum],rebspec,nbin)
	    }
	else {
	    if (debug) {
		call printf ("EIGENFIT2: rebinning %d-pixel spectrum\n")
		    call pargi (nbin)
		call flush (STDOUT)
		}
	    call rebin (Memr[spectrum],specsh, spshft,rebspec,nbin,minwav,dwav)
	    }
	if (debug) {
	    call printf ("EIGENFIT2: In  Lambda %f %f %f\n")
		call pargr (Memr[SX(specsh)])
		call pargr (Memr[SX(specsh)+1])
		call pargr (Memr[SX(specsh)+2])
	    call printf ("EIGENFIT2: In  Spectrum  %f %f %f\n")
		call pargr (Memr[SY(specsh)])
		call pargr (Memr[SY(specsh)+1])
		call pargr (Memr[SY(specsh)+2])
	    call printf ("EIGENFIT2: Out Lambda %f %f %f\n")
		call pargr (rebwl[1])
		call pargr (rebwl[2])
		call pargr (rebwl[3])
	    call printf ("EIGENFIT2: Out Spectrum %f %f %f\n")
		call pargr (rebspec[1])
		call pargr (rebspec[2])
		call pargr (rebspec[3])
	    }

	# Remove continuum and unwanted lines from object spectrum
	order = order1
	badproc = badproc1
        call eigencon (nbin, rebspec, rebwl, specname, nsmooth, ispec, nspec)

	# Initialize spectrum vector which will be filled with residuals
	do i = 1, nbin {
	    resspec1[i] = rebspec[i]
	    }

	# Fit first pass eigenvectors to object spectrum
	nfvec1 = nvec[1]
	if (nfvec1 > 0) {
	    call fitvecs (resspec1,nbin,rebvec1,nfvec1,coeff1,chi,chisq,FALSE)

	    # Compute actual sky removed
	    do i = 1, nbin {
		fitspec[i] = rebspec[i] - resspec1[i]
		}
	    if (debug) {
		call printf ("EIGENVEC2: returned from FITVECS pass 1\n")
		call flush (STDOUT)
		}
	    if (cfd1 > -1) {
		if (fcoef1) {
		    call fprintf (cfd1,"%s ")
			call pargstr (specfile)
		    }
		call fprintf (cfd1,"%4d %.3f %.3f")
		    call pargi (ispec)
		    call pargd (chi)
		    call pargd (chisq)
		do i = 1, nfvec1 {
		    call fprintf (cfd1," %.3f")
			call pargd (coeff1[i])
		    }
		call fprintf (cfd1,"\n")
		call flush (cfd1)
		}
	    if (debug) {
		call printf ("EIGENVEC2: coefficients saved\n")
		call flush (STDOUT)
		}

	    # Plot first pass residuals for this object spectrum
	    if (pltres1) {
		if (nspec > 0) {
		    call sprintf (title, SZ_LINE, "%d: %s object (pass 1 residuals)")
			call pargi (ispec)
			call pargstr (specname)
		    }
		else {
		    call sprintf (title, SZ_LINE, "%s object ( pass 1 residuals)")
			call pargstr (specname)
		    }
		call plotspec (nbin,resspec1,title,rebwl,xlab,nsmooth)
		if (nspec < 0 && pltspec) {
		    call sprintf (title, SZ_LINE, "%s pass 1 object+sky")
			call pargstr (specname)
		    call plotspec (nbin,rebspec,title,rebwl,xlab,nsmooth)
		    call sprintf (title, SZ_LINE, "%s pass 1 sky (fit)\n")
			call pargstr (specname)
		    call plotspec (nbin,fitspec,title,rebwl,xlab,nsmooth)
		    }
		}

	# Print results
	    if (debug) {
		call printf ("EIGENFIT2:  %d eigenvectors:\n")
		    call pargi (nfvec1)
		do i = 1, nfvec1 {
		    call printf ("EIGENFIT2:  %d: %g\n")
			call pargi (i)
			call pargr (coeff1[i])
		    }
		call printf ("EIGENFIT2: Res: %f %f %f\n")
		    call pargr (resspec1[1])
		    call pargr (resspec1[2])
		    call pargr (resspec1[3])
		}
		call flush (STDOUT)
	    }

	else {
	    call printf ("EIGENFIT2: No eigenvectors:  no fit\n")
	    call flush (STDOUT)
	    }

# Eignenvectors to which to fit residuals
	call clgstr ("eigenfile2",tempfiles,SZ_PATHNAME)
	call clgstr ("eigendir2",tempdir,SZ_PATHNAME)
	ldir = strlen (tempdir)
	if (ldir > 0 && tempdir[ldir] != '/' && tempdir[ldir] != '$') {
	    tempdir[ldir+1] = '/'
	    tempdir[ldir+2] = EOS
	    }

# Band (object, raw, sky) of eigenvector spectrum to use
	tband = clgeti ("eigenband2")

# Multispec eigenvector numbers (use only first if multiple files)
	call clgstr ("eigenum2", tempnums0, SZ_LINE)

# Get second pass eigenvectors
	minwav = minwav0
	maxwav = maxwav0
	call getvecs (tempfiles,tempdir,tband,tempnums0,minwav,maxwav,dwav,
		      spwl1, spwl2, rebvec2, nbin, rebwl, 2, nread)
	if (nread == 0)
	    go to endfit_

#  Process spectrum
	if (debug) {
	    call printf("EIGENFIT2: Overlap %9.4f to %9.4f by %9.4f, nbin = %d \n")
		call pargd (minwav)
		call pargd (maxwav)
		call pargd (dwav)
		call pargi (nbin)
	    call flush (STDOUT)
	    }

	#  2nd Eigenvectors are assumed to have the same wavelength range
	nfvec2 = 0
	if (nvec[2] > 0) {

	    # Initialize second residual vector to first residual vector
	    do i = 1, nbin {
		resspec2[i] = resspec1[i]
		}

	    # Remove continuum and unwanted lines from spectrum
	    badproc = badproc2
	    order = order2
            call eigencon (nbin,resspec2,rebwl,specname,nsmooth,ispec,nspec)

	    # Fit second pass eigenvectors to this spectrum
	    nfvec2 = nvec[2]
	    call fitvecs (resspec2,nbin,rebvec2,nfvec2,coeff2,chi,chisq,debug)
	    }
	else
	    call printf ("GETVECS: No second pass vectors read\n")

	if (nfvec2 > 0) {
	    if (debug) {
		call printf ("EIGENVEC2: returned from FITVECS pass 2\n")
		call flush (STDOUT)
		}

	    # Save second pass coefficients
	    if (cfd2 > -1) {
		if (fcoef1) {
		    call fprintf (cfd1,"%s ")
			call pargstr (specfile)
		    }
		call fprintf (cfd2,"%4d %.3f %.3f")
		    call pargi (ispec)
		    call pargd (chi)
		    call pargd (chisq)
		do i = 1, nfvec2 {
		    call fprintf (cfd2," %.3f")
			call pargd (coeff2[i])
		    }
		call fprintf (cfd2,"\n")
		}

	    do i = 1, nbin {
		fitspec[i] = rebspec[i] - resspec2[i]
		}

	    # Plot second pass residuals for this object spectrum
	    if (pltres2) {
		if (nspec > 0) {
		    call sprintf (title, SZ_LINE, "%d: %s object (pass 2 residuals)\n")
			call pargi (ispec)
			call pargstr (specname)
		    }
		else {
		    call sprintf (title, SZ_LINE, "%s object ( pass 2 residuals)\n")
			call pargstr (specname)
		    }
		call plotspec (nbin,resspec2,title,rebwl,xlab,nsmooth)
		if (nspec < 0 && pltspec) {
		    call sprintf (title, SZ_LINE, "%s pass 2 object+sky\n")
			call pargstr (specname)
		    call plotspec (nbin,rebspec,title,rebwl,xlab,nsmooth)
		    call sprintf (title, SZ_LINE, "%s pass 2 sky (fit)\n")
			call pargstr (specname)
		    call plotspec (nbin,fitspec,title,rebwl,xlab,nsmooth)
		    }
		}

	# Print results
	    if (debug) {
		call printf ("EIGENFIT2:  %d eigenvectors:\n")
		    call pargi (nfvec2)
		do i = 1, nfvec2 {
		    call printf ("EIGENFIT2:  %d: %g\n")
			call pargi (i)
			call pargr (coeff2[i])
		    }
		call printf ("EIGENFIT2: Res: %f %f %f\n")
		    call pargr (resspec2[1])
		    call pargr (resspec2[2])
		    call pargr (resspec2[3])
		}
	    }

	else
	    call printf ("EIGENFIT2: No eigenvectors fit on second pass\n")

#  Write final residual (object), raw, and sky spectra to file
	if (nspec < 0) {
	    call putspec (specim, specpath, specap, specband, imout, outpath,
		     resspec2, nbin, 1, 3)
	    call putspec (specim, specpath, specap, specband, imout, outpath,
		     rebspec, nbin, 2, 3)
	    call putspec (specim, specpath, specap, specband, imout, outpath,
		     fitspec, nbin, 3, 3)
	    }

#  Write residual (object) spectrum to new file or end of current file
	else {
	    call putspec (specim, specpath, specap, specband, imout, outpath,
		     resspec2, nbin, ispec, nspec)
	    }

# Write first pass continuum removal information to the header
	if (ispec == 1 && conproc1 != NOCONT) {
	    if (function == CN_SPLINE3)
		call strcpy (fstring, "spline3", SZ_LINE)
	    else if (function == CN_LEGENDRE)
		call strcpy (fstring, "legendre", SZ_LINE)
	    else if (function == CN_CHEBYSHEV)
		call strcpy (fstring, "chebyshev", SZ_LINE)
	    else
		call strcpy (fstring, "spline1", SZ_LINE)
	    call sprintf (tstring, SZ_LINE,
			  "Cont sub %s order %d, %d iter -%d %d")
		call pargstr (fstring)
		call pargi (order1)
		call pargi (niterate)
		call pargr (lowrej)
		call pargr (hirej)
	    call imastr (imout, "SVD1CONT", tstring);
	    }

# Write first pass coefficients to header
	if (nspec < 2) {
	    call imastr (imout, "SVD1FILE", tempfile)
	    if (debug) {
		call printf ("Eigenvector file %s\n")
		    call pargstr (tempfile)
		}
	    do ivec = 1, nfvec1 {
		call sprintf (keyword, 16, "SVD1C%03d")
		    call pargi (tmpspec[ivec])
		call imaddd (imout, keyword, coeff1[ivec])
		if (debug) {
		    call printf ("Coefficient %d:  vector %d: %f\n")
			call pargi (ivec)
			call pargi (tmpspec[ivec])
			call pargd (coeff1[ivec])
		    }
		}
		
	    }
	if (debug) {
	    call printf ("EIGENFIT2: First pass info written to image header\n")
	    call flush (STDOUT)
	    }

# Write second pass continuum removal information to the header
	if (ispec == 1 && conproc2 != NOCONT) {
	    if (function == CN_SPLINE3)
		call strcpy (fstring, "spline3", SZ_LINE)
	    else if (function == CN_LEGENDRE)
		call strcpy (fstring, "legendre", SZ_LINE)
	    else if (function == CN_CHEBYSHEV)
		call strcpy (fstring, "chebyshev", SZ_LINE)
	    else
		call strcpy (fstring, "spline1", SZ_LINE)
	    call sprintf (tstring, SZ_LINE,
			  "Cont sub %s order %d, %d iter -%d %d")
		call pargstr (fstring)
		call pargi (order)
		call pargi (niterate)
		call pargr (lowrej)
		call pargr (hirej)
	    call imastr (imout, "SVD2CONT", tstring);
	    }

# Write second pass coefficients to header
	if (nspec < 2) {
	    call imastr (imout, "SVD2FILE", tempfile)
	    if (debug) {
		call printf ("Pass 2 eigenvector file %s\n")
		    call pargstr (tempfile)
		}
	    do ivec = 1, nfvec2 {
		iv = nfvec1 + ivec
		call sprintf (keyword, 16, "SVD2C%03d")
		    call pargi (tmpspec[iv])
		call imaddd (imout, keyword, coeff2[ivec])
		if (debug) {
		    call printf ("Coefficient %d:  vector %d: %f\n")
			call pargi (ivec)
			call pargi (tmpspec[iv])
			call pargd (coeff2[ivec])
		    }
		}
		
	    }
	if (debug) {
	    call printf ("EIGENFIT2: Second pass info written to image header\n")
	    call flush (STDOUT)
	    }

# Close the object spectrum image
endfit_
	call close_image (specim, specsh)
	return
end

# May 18 2000	New subroutine, based on eigenfit.x

# Nov 20 2006	Call renormspec() instead of renorm()
