#*** File svdcomp.x
#*** By Mike Kurtz, modified by Doug Mink
#*** July 19, 1999


# Singular value decomposition of a bidiagonal matrix.
#	b = ub * s * vb**(t)
# where ub and vb are n by n orthogonal matrices and s is diagonal.
# If arguments v and c are n by n identity matrices, on exit they are
# replaced by vb and ub**t, respectively.

procedure svdcomp (d,e,n,v,iv,nrv,c,ic,ncc,ier)

double	d[n]		#  on input, d contains the diagonal elements
			# of the bidiagonal matrix b. d(i) = b(i,i),
			# i = 1,...,n.
			# on output, d contains the n (nonnegative)
			# singular values of b in nonincreasing order.

double	e[n]		# on input, e contains the superdiagonal
			# elements of b. e(1) is arbitrary,
			# e(i) = b(i-1,i), i=2,...,n.
			# on output, the contents of e are modified
			# by the subroutine.

int	n		# order of the matrix b. (input)

double	v[iv,1]		# nrv by n matrix. (input/output)
			# if nrv.le.0, v is not used. otherwise,
			# v is replaced by the nrv by n product
			# matrix v*vb. see remarks.

int	iv		# row dimension of matrix v exactly as
			# specified in the dimension statement in the
			# calling program. (input)

int	nrv		# number of rows of v. (input)

double	c[ic,1]		# n by ncc matrix. (input/output)
			# if ncc.le.0 c is not used. otherwise, c
			# is replaced by the n by ncc product
			# matrix ub**(t) * c. see remarks.

int	ic		# row dimension of matrix c exactly as
			# specified in the dimension statement in the
			# calling program. (input)

int	ncc		# number of columns in c. (input)

int	ier		# error parameter. (input)
			# warning error
			#   ier = 33 indicates that matrix b is not full
			#   rank or very ill-conditioned. small
			#	  singular values may not be very accurate.
			# terminal error
			#   ier = 129 indicates that convergence was
			#   not attained after 10*n qr sweeps.
			#   (convergence usually occurs in about 2*n sweeps).

# specifications for local variables
int	i,ii,j,k,kk,l,ll,lp1,nqrs,n10
bool	wntv,havers,fail
double	dnorm,zero,one,two,cs,f,g,h,sn,t,x,y,z

define	testsplit_	10

begin
	zero = 0.d0
	one = 1.d0
	two = 2.d0
	ier = 0
	if (n < 0)
	    return
	n10 = 10 * n
	if (nrv > 0)
	   wntv = TRUE
	else
	   wntv = FALSE
	if (ncc > 0)
	    havers = TRUE
	else
	    havers = FALSE
	fail = FALSE
	nqrs = 0
	e[1] = zero
	dnorm = zero
	do j = 1, n {
	    dnorm = max (abs(d[j])+abs(e[j]), dnorm)
	    }

	do kk = 1, n {
	    k = n + 1 - kk
#                                  test for splitting or rank
#                                    deficiencies first make test for
#                                    last diagonal term, d(k), being
#                                    small.
testsplit_
   	    if (k != 1) {
	        t = dnorm + d[k]

# Since d(k) is small we will make a special pass to transform e(k) to zero.
		if (t == dnorm) {
		    cs = zero
		    sn = -one
		    do ii = 2, k {
			i = k + 1 - ii
			f = -sn * e[i+1]
			e[i+1] = cs * e[i+1]
			t = d[i]
			call lsvg1 (t,f,cs,sn,d[i])

# Transformation constructed to zero position (i,k).
			if (wntv) {
			    do j = 1,nrv {
				call lsvg2 (cs,sn,v(j,i),v(j,k))
				}
			    }
			}
		    }
		}

# Accumulate rt. transformations in v.  the matrix is now bidiagonal, and of
# lower order since e(k) = zero
	    do ll = 1, k { 
		l = k + 1 - l 
		t = dnorm + e[l]
		if (t == dnorm)
		    break
		t = dnorm+d[l-1]
		if (t == dnorm) {
		    cs = zero
		    sn = -one
		    do i = l, k {
			f = -sn * e[i]
			e[i] = cs * e[i]
			t = dnorm + f
			if (t == dnorm)
			    break
			t = d[i]
			call lsvg1 (t,f,cs,sn,d[i])
			if (havers) {
			    do j = 1,ncc {
				call lsvg2 (cs,sn,c[i,j],c[l-1,j])
				}
			    }
			}
		     break
		     }
		}

# This loop cannot complete since e(1) = zero.
# cancellation of e(l), l > 1.  test for convergence
	    z = d[k]

# Shift from bottom 2 by 2 minor of b**(t)*b.
	    if (l != k) {
		x = d[l]
		y = d[k-1]
		g = e[k-1]
		h = e[k]
		f = ((y-z)*(y+z) + (g-h)*(g+h)) / (two*h*y)
		g = sqrt (one+f**2)
		if (f < zero)
		    t = f - g
		else
		    t = f + g
		f = ((x-z)*(x+z) + h*(y/t-h)) / x
#                                  next qr sweep
		cs = one
		sn = one
		lp1 = l+1
		do i = lp1, k {
		    g = e[i]
		    y = d[i]
		    h = sn * g
		    g = cs * g
		    call lsvg1 (f,h,cs,sn,e[i-1])
		    f = (x * cs) + (g * sn)
		    g = (-x * sn) + (g * cs)
		    h = y * sn
		    y = y * cs

#		accumulate rotations (from the right) in v
		    if (wntv) {
			do j = 1, nrv {
			    call lsvg2 (cs,sn,v(j,i-1),v(j,i))
			    }
			}

		    call lsvg1 (f,h,cs,sn,d(i-1))
		    f = cs*g+sn*y
		    x = -sn*g+cs*y

		    if (havers) {
			do j = 1,ncc {
			    call lsvg2 (cs,sn,c(i-1,j),c(i,j))
			    }
			}
		    }

	    # Apply rotations from the left to right hand sides in c
		e[l] = zero
		e[k] = f
		d[k] = x
		nqrs = nqrs + 1
		if (nqrs < n10) go to testsplit_
#                                  return to test for splitting.
		fail = TRUE
		}
# Cutoff for convergence failure. nqrs will be 2*n usually.
	    if (z < zero) {
		d[k] = -z
		if (wntv) {
		    do j = 1, nrv {
			v[j,k] = -v[j,k]
			}
		    }
		}
	    }

# Convergence. d[k] is made nonnegative
	if (n != 1) {

	# Check to see whether singular values are in order
	    do i = 2,n {
		if (d[i] > d[i-1])
		    break
		}

	# Not every singular value is in order
	    if (i <= n) {

		do i = 2,n {
		    t = d[i-1]
		    k = i - 1
		    do j = i, n {
			if (t < d[j]) {
			    t = d[j]
			    k = j
			    }
			}

		    if (k != i-1) {
			d[k] = d[i-1]
			d[i-1] = t
			if (havers) {
			    do j = 1, ncc {
				t = c[i-1,j]
				c[i-1,j] = c[k,j]
				c[k,j] = t
				}
			    }
			if (wntv) {
			    do j = 1, nrv {
				t = v[j,i-1]
				v[j,i-1] = v[j,k]
				v[j,k] = t
				}
			    }
			}
		    }
		}
	    }
#                                  end of ordering algorithm.

	ier = 129

#  Check for possible rank deficiency
	if (!fail) {
	    ier = 33
	    t = 0.d0
	    if (d[1] != zero)
		t = d[n] / d[1]
	    f = 100.d0 + t
	    if (f != 100.d0)
		ier = 0
	    }
	return
end


#  LSVDF - singular value decomposition of a double matrix

procedure lsvdf  (a,ia,m,n,b,ib,nb,s,wk,ier)

double	a[ia,n]	# real m by n matrix. (input/output)
		# on input, a contains the matrix to be decomposed.
		# on output, a contains the n by n matrix v
		# in its first n rows. see remarks.
		# Either m >= n or m < n is permitted.

int	ia	# Row dimension of matrix a exactly as specified in the
		# dimension statement in the calling program. (input)
		# a is used by lsvdf as work storage for an n by n matrix.
		# Therefore, ia must be greater than or equal to max(m,n).

int	m	# Number of rows in a. (input)

int	n	# Number of columns in a. (input)

double	b[ib,1]	# m by nb matrix. (input/output)
		# b is not used if nb <= 0. Otherwise, b is replaced by
		# the matrix product u**(t) * b.  see remarks.

int	ib	# Row dimension of matrix b exactly as specified in the
		# dimension statement in the calling program. (input)

int	nb	# Number of columns in b. (input)
		# if nb <= 0, b is not used.

double	s[n]	# vector of length n. (output)
		# on output, s contains the ordered singular values of a.
		# s[1] >= s[2],..., >= s[n] >= 0.

double	wk[n,2]	# Work vector of length 2n.

int	ier	# error parameter. (output)
		# Warning errors:
		#   ier = 33 indicates that matrix a is not full rank or very
		#	     ill-conditioned. small singular values may not be
		#	     very accurate.
		#   ier = 34 indicates that either n <= 0 or m <= 0.
		# Terminal error:
		#   ier = 129 indicates that convergence was not obtained by
		#	     svdcomp and computation was discontinued.
#
#   remarks  1.  lsvdf computes the singular value decomposition of
#                a real m by n matrix
#                     a = u * q * v**(t)  where
#                u is an m by m orthogonal matrix,
#                v is an n by n orthogonal matrix, and
#                q is an m by n matrix with all elements zero except
#                     q(i,i) = s(i) i=1,...,min(m,n).
#                v is returned in the first n rows of a.
#                u is obtained by setting b to the m by m identity
#                matrix, on input, and setting nb = m. on output, b is
#                replaced by u**(t).
#            2.  the notation u**(t) and v**(t) represents u
#                transpose and v transpose, respectively. q**(+)
#                denotes the generalized inverse of q.
#            3.  lsvdf is useful in analyzing and solving the least
#                squares problem a*x.appr.b (where .appr. means
#                approximately equals). in this case b is a vector of
#                length m and lsvdf is called with ib = m, nb=1. the
#                solution is x = v*q**(+)*u**(t)*b. u**(t)*b replaces
#                b on output. the solution x is obtained as follows...
#                (the user may wish to set small singular values such as
#                s(i) to zero just prior to computing x. see reference
#                for details.)
#
#                c                  compute q**(+) * u**(t) * b
#                      l = min0(m,n)
#                      do 10 i = 1,l
#                         t = 0.0
#                         if (s(i).ne.0.0) t = b(i)/s(i)
#                         b(i) = t
#                   10 continue
#                c                  compute v * q**(+) * u**(t) * b
#                   15 do 25 i = 1,n
#                         x(i) = 0.0
#                         do 20 j = 1,l
#                   20    x(i) = x(i)+a(i,j)*b(j)
#                   25 continue
#                if b is set to the j-th column of the m by m identity
#                matrix on input, x is the j = th column of a**(+) (the
#                generalized inverse of a).
#            4.  the user should be aware of several practical aspects
#                of the singular value analysis. one of these is the
#                effect of the uncertainty of the data and the need for
#                scaling. see the lawson-hanson reference pages 180-198
#                for details.

int	i,j,jp1,k,l,mm,nn,nnp1,ns,nsp1
int	level, nm, i0
double	f
double	zero,one,t
bool	allzero

begin
	zero = 0.d0
	one = 1.d0
	ier = 0

	nn = n
	ier = 34
	if (nn <= 0 || m <= 0)
	    return
	ier = 0

# for zero rows and cols. pack the nonzero cols to the left
	do j = n, 1, -1 {
	    allzero = TRUE
	    do i = 1, m {
		if (a[i,j] != zero)
		    allzero = FALSE
		}

	    # if col j is zero. exchange it with col n
	    if (allzero) {
		if (j < nn) {
		    do i = 1,m {
			a[i,j] = a[i,nn]
			}
		    }
		a[1,nn] = j
		nn = nn - 1
		}
	    }

# If n = 0 then a is entirely zero and svd computation can be skipped
	ns = 0
	if (nn > 0) {

	# Pack nonzero rows to the top; quit packing if find n nonzero rows
	    i = 1
	    mm = m
	    do i0 = 1, n {

		if (a[i,i] == zero) {
		    allzero = TRUE
		    do j = 1,nn {
			if (a[i,j] != zero)
			    allzero = FALSE
			}

		    # If row i is zero, exchange rows i and m
		    if (allzero) {
			if (nb > 0) {
			    do j = 1, nb {
				t = b[i,j]
				b[i,j] = b[mm,j]
				b[mm,j] = t
				}
			    }

			do j = 1,nn {
			    a[i,j] = a[mm,j]
			    }
			if (mm <= nn) {
			    do j = 1,nn {
				a[mm,j] = zero
				}
			    }

			# exchange is finished
			mm = mm - 1
			}
		    else
			i = i + 1
		    }
		}

# End special for zero rows and columns
# Begin svd algorithm..
# (1) Reduce the matrix to upper bidiagonal form with householder
#     transformations.
#     h(n)...h(1)aq(1)...q(n-2) = (d**t,0)**t where d is upper bidiagonal.
# (2) Apply h(n)...h(1) to b. here h(n)...h(1)*b replaces b in storage.
# (3) The matrix product w = q(1)...q(n-2) overwrites the first n rows of a
#     in storage.
# (4) An SVD for d is computed. here k rotations ri and pi are computed
#     so that rk...r1*d*p1**(t)...pk**(t) = diag(s1,...,sm) to working
#     accuracy. the si are nonnegative and nonincreasing. here rk...r1*b
#     overwrites b in storage while a*p1**(t)...pk**(t) overwrites a
#     in storage.
# (5) It follows that,with the proper definitions, u**(t)*b overwrites
#     b, while v overwrites the first n row and columns of a.

	    l = min0 (mm,nn)

# The following loop reduces a to upper bidiagonal and also applies
# the premultiplying transformations to b.
	    do j = 1, l {
		if (j < mm) {
		    jp1 = min0 (j+1,nn)
		    call vhs12 (1,j,j+1,mm,a[1,j],1,t,a[1,jp1],1,ia,nn-j)
		    call vhs12 (2,j,j+1,mm,a[1,j],1,t,b,1,ib,nb)
		    }
		if (j < nn-1) {
		    call vhs12 (1,j+1,j+2,nn,a[j,1],ia,wk[j,2],a[j+1,1],ia,1,mm-j)
		    }
		}

# Copy the bidiagonal matrix into the array s for svdcomp
	    if (l != 1) {
		do j = 2, l {
		    s[j] = a[j,j]
		    wk[j,1] = a[j-1,j]
		    }
		}

	    s[1] = a[1,1]

	    ns = nn
	    if (mm < nn) {
		ns = mm + 1
		s[ns] = zero
		wk[ns,1] = a[mm,mm+1]
		}

	# Construct the explicit n by n product matrix, w = q1*q2*...*ql*i
	# in the array a
	    do k = 1, nn {
		i = nn + 1 - k
		if (i <= min0 (mm,nn-2))
		    call vhs12 (2,i+1,i+2,nn,a[i,1],ia,wk[i,2],a[1,i+1],1,ia,nn-i)
		do j = 1, nn {
		    a[i,j] = zero
		    }
		a[i,i] = one
		}

	# Compute the svd of the bidiagonal matrix
	    level = 1
	    call svdcomp (s[1],wk[1,1],ns,a,ia,nn,b,ib,nb,ier)

	    if (ier > 128)
		return

	# Test for ier = 33
	    if (ier == 33) {
		t = 0.d0
		nm = min0(m,n)
		if (s[1] != zero)
		    t = s[nm]/s[1]
		f = 100.d0 + t
		if (f != 100.d0)
		    ier = 0
		}
	    }

	if (ns < nn) {
	    nsp1 = ns+1
	    do j = nsp1,nn {
		s[j] = zero
		}
	    }

	# Special for zero rows and columns
	if (nn != n) {
	    nnp1 = nn+1

	# Move record of permutations and store zeros
	    do j = nnp1,n {
		s[j] = a[1,j]
		if (nn >= 1) {
		    do i = 1,nn {
  			a[i,j] = zero
			}
		    }
		}

	    # Permute rows and set zero singular values
	    do k = nnp1,n {
		i = Idnint (s[k])
		s[k] = zero
		do j = 1,n {
		    a[k,j] = a[i,j]
		    a[i,j] = zero
		    }
		a[i,k] = one
		}
	    }

	return
end

#   LSVG1 - nucleus called only by routine LSVDB

procedure lsvg1  (a,b,cs,sn,sig)

double	a,b,cs,sn,sig

double	aa,bb

begin
	if (abs(a) > abs(b)) {
	    aa = abs (a + a)
	    sig = aa * sqrt (0.25d0 + (b / aa)**2)
	    cs = a / sig
	    sn = b / sig
	    }
	else if (b != 0.d0) {
	    bb = abs (b + b)
	    sig = bb * sqrt (0.25d0 + (a / bb)**2)
	    cs = a / sig
	    sn = b / sig
	    }
	else {
	    sig = 0.d0
	    cs = 0.d0
	    sn = 1.d0
	    }
	return
end


# LSVG2 - nucleus called only by routine svdcomp

procedure lsvg2  (cs,sn,x,y)

double	cs,sn,x,y

double	xr

begin
	xr = (cs * x) + (sn * y)
	y = (-sn * x) + (cs * y)
	x = xr
	return
end


# DROTG - construct givens plane rotation

	procedure drotg  (da,db,dc,ds)

double	da	# first element of double vector.  (input/output)
		# on output, r = (+/-)dsqrt(da**2 + db**2) overwrites da.

double	db	# second element of double vector.  (input/output)
		# on output, z overwrites db.
		# z is defined to be:
		#   ds,       if dabs(da) > dabs(db)
		#   1.0d0/dc, if dabs(db) >= dabs(da) and dc != 0.0d0
		#   1.0d0,    if dc == 0.0d0.

double	dc	# double element of output transformation matrix. see remarks.

double	ds	# double element of output transformation matrix. see remarks.

#  DROTG constructs the givens transformation
#          ( dc  ds )
#      g = (        ) ,  dc^2 + ds^2 = 1 ,
#          (-ds  dc )
#  which zeros the second element of (da,db)**t.

double   u,v,r

begin

	if (dabs(da) > dabs(db)) {
	    u = da + da		# note that u and r have the sign of da
	    v = db / u
	    r = dsqrt (.25d0 + v**2) * u
	    dc = da / r		# note that dc is positive
	    ds = v * (dc + dc)
	    db = ds
	    da = r
	    }

	else if (db != 0.d0) {
	    u = db + db		# note that u and r have the sign of db
	    v = da / u

	    # r is immediately stored in da
	    da = dsqrt (.25d0 + v**2) * u
	    ds = db / da		# note that ds is positive
	    dc = v * (ds + ds)
	    if (dc == 0.d0)
		db = 1.d0
	    else
		db =  1.d0 / dc
	    }

# here da = db = 0.d0
	else {
	    dc = 1.d0
	    ds = 0.d0
	    }
	return
end


#  VHSL2 - real householder transformation - computation and applications.

procedure vhs12  (mode,lp,l1,m,u,incu,up,c,incc,icv,ncv)

int	mode		# option parameter. (input)
			# if mode = 1, the subroutine computes a
			#    householder transformation and if ncv > 0,
			#    multiplies it by the set of ncv vectors
			#    (each of length m) stored in c. for a
			#    given vector v of length m and two int
			#    indices lp and l1 that satisfy
			#    1 .le. lp .lt. l1 .le. m, the subroutine
			#    defines an m by m householder
			#    transformation q which satifies qv = w where
			#    w(i) = v(i) for i.lt.lp
			#    w(lp) = -sig*sqrt(v(lp)**2+v(l1)**2+...
			#      +v(m)**2)
			#      sig = 1  if v(lp).ge.0
			#      sig = -1 if v(lp).lt.0
			#    w(i) = v(i) for lp.lt.i.lt.l1
			#    w(i) = 0    for i.ge.l1.
			# if mode = 2, the subroutine assumes that a
			#    householder transformation has already been
			#    defined by a previous call with mode = 1, and
			#    if ncv > 0, multiplies it by the set of ncv
			#    vectors (each of length m) stored in c.

int	lp		# parameters that define the desired householder
int	l1		# transformation. (input)
int	m		# if the condition 1.le.lp.lt.l1.le.m is not
			# satisfied, the subroutine returns to the calling
			# program without performing any computations.

double	u[ARB]		# vector of m elements. (input, and output if mode = 1)
			# The storage increment between elements of u is incu.
			# (i.e., u(1+(j-1)*incu), j = 1,...,m).
			# If mode=1, the array v is defined as
			# v(j) = u(1+(j-1)*incu), j = 1,...,m.
			# On output, u(1+(lp-1)*incu) is set to w(lp)
			# (as defined above in the description of mode = 1).

int	incu		# increment between elements of u. (input)

double	up		# scalar set to v(lp)-w(lp) to define the householder
			# transformation q.
			# (input if mode = 2, output if mode=1)

double	c[ARB]		# vector of ncv*m elements. (input/output)
			# if ncv <= 0, c is not used.
			# if ncv > 0, c contains ncv vectors of length
			# m with increment incc between elements of vectors
			# and increment icv between vectors. element i of
			# vector j is defined as c(1+(i-1)*incc+(j-1)*icv),
			# i = 1,...,m and j=1,...,ncv.
			# On output, c contains the set of ncv vectors
			# resulting from multiplying the given vectors by q.

int	incc		# increment between elements of vectors in c. (input)

int	icv		# increment between vectors in c. (input)

int	ncv		# number of vectors stored in c. (input)

#   remarks  1.  if u is a single subscripted array or the j-th column
#                of a matrix, then incu = 1. if u is the i-th row of a
#                matrix then incu is the row dimension of the matrix
#                exactly as specified in the calling program.
#            2.  if c is a double subscripted matrix and the vectors
#                are the first ncv columns of c, then incc = 1 and icv
#                is the row dimension of c exactly as specified in
#                the calling program. in this case c is replaced
#                by qc. if the vectors are successive rows of c
#                then incc is the row dimension of c exactly as
#                specified in the calling program and icv = 1. in this
#                case c is replaced by cq.

int	ij,ilp,il1,im,incr,i2,i3,i4,j
double	sm,b
double	one,cl,clinv,sm1

begin

	one = 1.d0

	if (0 >= lp || lp >= l1 || l1 > m)
	    return
	ilp = (lp - 1) * incu + 1
	il1 = (l1 - 1) * incu + 1
	im = (m - 1) * incu + 1
	cl = abs (u[ilp])

#  Construct the transformation.
	if (mode == 2)
	    if (cl <= 0.d0) return
	else {
	    do ij = il1, im, incu {
		cl = max(abs(u(ij)),cl)
		}
	    if (cl <= 0.d0)
		return
	    clinv = one / cl
	    sm = (double (u[ilp]) * clinv)**2
	    do ij = il1,im,incu {
		sm = sm + (double (u[ij]) * clinv)**2
		}

	    # convert double precision sm to single precision sm1
	    sm1 = sm
	    cl = cl * sqrt(sm1)
	    if (u[ilp] > 0.d0) cl = -cl
	    up = u[ilp]-cl
	    u[ilp] = cl
	    }

# Apply the transformation i+u*(u**t)/b to c.
	if (ncv <= 0) return

# b must be nonpositive here. if b = 0., return.
	b = double (up) * u[ilp]
	if (b >= 0.d0) return
	b = one / b
	i2 = 1 - icv + (incc * (lp-1))
	incr = incc * (l1 - lp)
	do j = 1,ncv {
	    i2 = i2 + icv
	    i3 = i2 + incr
	    i4 = i3
	    sm = c[i2] * double (up)
	    do ij = il1, im, incu {
		sm = sm + c[i3] * double (u[ij])
		i3 = i3 + incc
		}
	    if (sm != 0.d0) {
		sm = sm * b
		c[i2] = c[i2] + sm * double (up)
		do ij = il1, im, incu {
		    c[i4] = c[i4] + sm * double (u[ij])
		    i4 = i4 + incc
		    }
		}
	    }

	return
end
