# File svdfit/Src/t_svdres.x
# February 5, 2009
# By Doug Mink, Harvard-Smithsonian Center for Astrophysics
 
# SVDRES is an IRAF task for fitting eigenvectors to a spectrum and returning
# the residuals as a spectrum file.
# For arguments, see parameter file svdres.par.
# Information is shared in common blocks defined in "rvsao.com".

# Copyright(c) 2000-2009 Smithsonian Astrophysical Observatory
# You may do anything you like with this file except remove this copyright.
# The Smithsonian Astrophysical Observatory makes no representations about
# the suitability of this software for any purpose.  It is provided "as is"
# without express or implied warranty.
 
include	<imhdr.h>
include	<imio.h>
include	<fset.h>
include	<smw.h>
include "rvsao.h"
include "contin.h"
include "emv.h"
 
define	LEN_USER_AREA	100000
define	EIG_OUTFORM	"|none|oned|twod|multispec|"
define	EIG_NONE	1
define	EIG_ONED	2
define	EIG_TWOD	3
define	EIG_MULTISPEC	4

procedure t_svdres ()

int	i
char	specfiles[SZ_PATHNAME]	# Object spectrum file names
char	specfile[SZ_PATHNAME]	# Object spectrum file name
char	specpath[SZ_PATHNAME]	# Object spectrum path name
char	specdir[SZ_PATHNAME]	# Directory for object spectra
char	resfile[SZ_PATHNAME]	# Output spectrum file name
char	respath[SZ_PATHNAME]	# Output spectrum path name
char	resdir[SZ_PATHNAME]	# Directory for output spectra
char	coeffile[SZ_PATHNAME]	# Output fit coefficients
int	rmode			# Report format (1=normal,2=one-line)
int	logfiles		# List of log files
char	logfile[SZ_PATHNAME] # Log file name
char	wtitle[20]		# Title for wavelength plots of spectrum
int	specap			# Object aperture to read from multispec file
int	specband		# Object band to read from multispec file
pointer	speclist		# List of spectrum files
pointer	specim			# Image header etc.
char	str[SZ_LINE]
int	fd
char	outform[SZ_LINE]	# Output file format
char	contstr[SZ_LINE]	# Remove continuum
				# (no subtract divide)
int	oform
int	nspecap,nspecap0	# Number of object multispec spectra
int	specap_range[3,MAX_RANGES]
int	ip,jp,lfile		# Limits for multispec aperture decoding
char	lbracket[3]		# "[({"
char	rbracket[3]		# "])}"
pointer	rebspec			# Allocated vector for rebinned spectra
pointer	rebwl			# Allocated vector for rebinned wavelengths
pointer	resspec			# Allocated vector for spectra residuals
pointer	rebcont			# Allocated vector for spectra continuua
pointer	reblines		# Allocated vector for lines removed from spectra
pointer	fitspec			# Allocated vector for removed eigenvector fit
int	ispec, nspec
int	lname, lroot
int	nbin
pointer	imout
char	jdot
char	msfx[8]
bool	plotint
int	iindef
int	cfd
int	ivec
char	fstring[SZ_LINE]
char	rastr[16], decstr[16]
bool	verbose
bool	aptrack
double	fitmax
 
bool	clgetb()
int	clpopnu(), clgeti(), clgfil(), open()
int	strdic(), stridx(), stridxs(), strldx(), strmatch()
int	decode_ranges(),get_next_number()
int	imtgetim(), imaccess(), strlen(), ldir, access()
double	clgetd()
pointer	imtopen()
pointer	immap()
errchk	immap()

define	newspec_ 10
define	newap_	 20
define	endeig_	 90

include	"rvsao.com"
include	"contin.com"
include	"emv.com"
 
begin
	c0 = 299792.5
	nfound = 0
	nctemp = 0
	ispec = 0
	imout = NULL
	do ivec = 1, MAXTEMPS {
	    tmpsh[ivec] = NULL
	    tmpim[ivec] = NULL
	    rebvec[ivec] = NULL
	    cnvec[ivec] = NULL
	    }
	rebspec = NULL
	rebwl = NULL
	resspec = NULL
	rebcont = NULL
	reblines = NULL
	fitspec = NULL
	iindef = INDEFI

	jdot = '.'
	call strcpy (".ms.imh", msfx, 8)
	msfx[1] = jdot
	msfx[4] = jdot

	call sprintf (lbracket,3,"[({")
	call sprintf (rbracket,3,"])}")
	call sprintf (wtitle,20,"Wavelength")

# Get task parameters.

# Spectra to cross-correlate
	call clgstr ("spectra", specfiles,SZ_PATHNAME)

# Multispec spectrum numbers (use only first if multiple files)
	call clgstr ("specnum",specnums,SZ_LINE)
	if (decode_ranges (specnums, specap_range, MAX_RANGES, nspecap0) == ERR){
	    call sprintf (str, SZ_LINE, "SVDRES: Illegal multispec list <%s>")
		call pargstr (specnums)
	    call error (1, str)
	    }
	call clgstr ("specdir",specdir,SZ_PATHNAME)
	ldir = strlen (specdir)
	if (specdir[1] != EOS && specdir[ldir] != '/') {
	    specdir[ldir+1] = '/'
	    specdir[ldir+2] = EOS
	    }
	specband = clgeti ("specband")

# Optional intermediate data plot switches
	pltspec = clgetb ("obj_plot")
	pltcon  = clgetb ("contsub_plot")
	plttemp = clgetb ("vec_plot")

# Print processing information
	debug  = clgetb ("debug")
	verbose  = clgetb ("verbose")

# Fit continuum to input spectra
	call clgstr ("cont_remove",contstr,SZ_LINE)
	conproc = strdic (contstr, contstr, SZ_LINE, CONT_TYPE)
	if (debug) {
	    call printf ("SVDRES: Input spectra continuum ")
	    if (conproc == DIVCONT)
		call printf ("divided\n")
	    else if (conproc == SUBCONT)
		call printf ("subtracted before processing\n")
	    else if (conproc == OUTCONT)
		call printf ("subtracted after processing\n")
	    else if (conproc == ZEROCONT)
		call printf ("divided - 1\n")
	    else
		call printf ("not removed\n")
	    }

# Fit continuum to eigenvectors
	call clgstr ("vec_cont",contstr,SZ_LINE)
	convec = strdic (contstr, contstr, SZ_LINE, CONT_TYPE)
	if (debug) {
	    call printf ("SVDRES: Eitenvector continuum ")
	    if (convec == DIVCONT)
		call printf ("divided\n")
	    else if (convec == SUBCONT)
		call printf ("subtracted before processing\n")
	    else if (convec == OUTCONT)
		call printf ("subtracted after processing\n")
	    else if (convec == ZEROCONT)
		call printf ("divided - 1\n")
	    else
		call printf ("not removed\n")
	    }

# Continuum fit parameter pset
	if (conproc != NOCONT || convec != NOCONT)
	    call cres_get_pars()

# Number of times to smooth (1-2-1) final data plot
	nsmooth = clgeti ("nsmooth")

# Interact with graphics displays
        plotint = clgetb ("plot_int")
        if (plotint)
            nsmooth = nsmooth
        else
            nsmooth = -(nsmooth + 1)

# Format of output file
	call clgstr ("res_format",outform,SZ_LINE)
	oform = strdic (outform,outform,SZ_LINE, EIG_OUTFORM)
	if (debug) {
	    call printf ("SVDRES: output file is type %d\n")
		call pargi (oform)
	    call flush (STDOUT)
	    }
	if (oform == EIG_NONE)
	    nspec = 0
	else if (oform == EIG_ONED)
	    nspec = 1
	else if (oform == EIG_MULTISPEC)
	    nspec = -3
	else {
	    call countspec (specfiles, specdir, specnums, specband, debug, nspec)
	    }

# Report mode for log file
	rmode = 1
	rmode = clgeti ("report_mode")

# Number of points into which to rebin spectrum and eigenvectors
	nbin = iindef
	nbin = clgeti ("nbin")
	if (debug) {
	    call printf ("SVDRES: nbin = %d\n")
		call pargi (nbin)
	    }

	contvec = NULL

# Eigenvector tracking parameter
	aptrack = FALSE
	aptrack = clgetb ("eigentrack")

# Initialize emission and absorption lines for labelling
	call eminit (FALSE)

# Bad line removal
        call clgstr ("fixbad", fstring, SZ_LINE)
	badproc = strdic (fstring, fstring, SZ_LINE, BAD_TYPE)
	if (debug) {
	    call printf ("SVDRES: Input spectra bad lines ")
	    if (badproc == FILLBAD)
		call printf ("filled\n")
	    else if (badproc == ZEROBAD)
		call printf ("zeroed\n")
	    else if (badproc == NZEROBAD)
		call printf ("filled then zeroed\n")
	    else
		call printf ("not removed\n")
	    }

# Fitting threshold for single vector fit (such as sky removal)
	fitmax = 2.0
	fitmax = clgetd ("fitmax")
 
# Open log files and write a header.
	logfiles = clpopnu ("logfiles")
	call fseti (STDOUT, F_FLUSHNL, YES)
	i = 0
	call strcpy ("svdfit.svdres",taskname,SZ_LINE)

	if (verbose) {
	    while (clgfil (logfiles, logfile, SZ_PATHNAME) != EOF) {
		fd = open (logfile, APPEND, TEXT_FILE)
		if (fd == ERR) break
		if (rmode == 1) {
		    call loghead (taskname,str)
		    call fprintf (fd, "%s\n")
			call pargstr (str)
		    }
		i = i + 1
		logfd[i] = fd
		}
	    nlogfd = i
	    call clpcls (logfiles)
	    }

# Spectra to cross-correlate
	speclist = imtopen (specfiles)

# Get next object spectrum file name from the list
newspec_
	if (imtgetim (speclist, specfile, SZ_PATHNAME) == EOF)
	   go to endeig_
#	if (debug) {
#	    call printf ("SVDRES: next file is %s\n")
#		call pargstr (specfile)
#	    }

# Check for specified apertures in multispec spectrum file
	ip = stridxs (lbracket,specfile)
	if (ip > 0) {
	    lfile = strlen (specfile)
	    specfile[ip] = EOS
	    jp = 0
	    ip = ip + 1
	    while (stridx (specfile[ip],rbracket) == 0 && ip <= lfile) {
		jp = jp + 1
		specnums[jp] = specfile[ip]
		specfile[ip] = EOS
		ip = ip + 1
		}
	    if (jp > 0)
		specnums[jp+1] = EOS
	    else
		call strcpy ("0",specnums,SZ_LINE)
	    if (decode_ranges (specnums,specap_range,MAX_RANGES,nspecap) == ERR){
		call sprintf (str, SZ_LINE, "SVDRES: Illegal multispec list <%s>")
		    call pargstr (specnums)
		call error (1, str)
		}
	    }
	else
	    nspecap = nspecap0
	if (debug) {
	    call printf ("SVDRES: next file is %s [%s] = %d aps\n")
		call pargstr (specfile)
		call pargstr (specnums)
		call pargi (nspecap)
	    }

# Check for readability of object spectrum
	call strcpy (specdir,specpath,SZ_PATHNAME)
	call strcat (specfile,specpath,SZ_PATHNAME)
	if (imaccess (specpath, READ_ONLY) == NO) {
	    call eprintf ("SVDRES: cannot read spectrum file %s \n")
		call pargstr (specpath)
	    go to newspec_
	    }

# Allocate rebinned, wavelength, residual, and sky spectrum buffers
	if (rebspec == NULL) {
	    if (nbin == iindef || nbin == 0) {
		specim = NULL
		iferr (specim = immap (specpath, READ_ONLY, LEN_USER_AREA)) {
		    if (specim != NULL) call imunmap (specim)
		    call eprintf ("GETIMAGE: Cannot read image %s\n")
			call pargstr (specpath)
		    specim = ERR
		    return
		    }

		call imgipar (specim, "i_naxis1", nbin)
		call imgspar (specim, "RA", rastr, 15)
		call imgspar (specim, "DEC", decstr, 15)
		if (debug) {
		    call printf ("SVDRES: nbin = %d, %s %s\n")
			call pargi (nbin)
			call pargstr (rastr)
			call pargstr (decstr)
		    }
		if (specim != NULL) call imunmap (specim)
		specim = NULL
		}
	    call malloc (rebspec, nbin, TY_REAL)
	    call malloc (rebwl, nbin, TY_REAL)
	    call malloc (resspec, nbin, TY_REAL)
	    call malloc (rebcont, nbin, TY_REAL)
	    call malloc (reblines, nbin, TY_REAL)
	    call malloc (fitspec, nbin, TY_REAL)
	    }

# Get or create name of output file
	call clgstr ("resfile",resfile,SZ_PATHNAME)
	lfile = strlen (resfile)
	if (lfile == 0) {
	    lname = strmatch (specfile, msfx)
	    if (lname > 0) {
		lname = lname - strlen (msfx)
		lroot = lname - 1
		if (nspec > 0)
		    lname = lname + 3
		}
	    if (lname == NULL) {
		lname = strldx (jdot, specfile)
		lroot = lname - 1
		}
	    if (lname == NULL)
		lroot = lfile
	    call strcpy (specfile, resfile, lroot)
	    call strcat ("r", resfile, SZ_PATHNAME)
	    if (lname != NULL)
		call strcat (specfile[lname], resfile, SZ_PATHNAME)
	    }

# Get name of output directory
	call clgstr ("resdir",resdir,SZ_PATHNAME)
	ldir = strlen (resdir)

# Create directory if it does not already exist
	if (resdir[1] != EOS) {
	    if (resdir[ldir] != '/') {
		ldir = ldir + 1
		resdir[ldir+1] = EOS
		}
	    resdir[ldir] = EOS
	    if (access (resdir, 0, 0) == NO)
		call fmkdir (resdir)
	    resdir[ldir] = '/'
	    }

# Check for writability of output spectrum
	call strcpy (resdir,respath,SZ_PATHNAME)
	call strcat (resfile,respath,SZ_PATHNAME)

# Open file in which to save coefficients
	cfd = -1
	call strcpy (respath, coeffile, SZ_PATHNAME)
	call strcat (".coeff", coeffile, SZ_PATHNAME)
	if (nspec > 1) {
	    cfd = open (coeffile, NEW_FILE, TEXT_FILE)
	    if (cfd < 2) {
		call eprintf ("SVDRES: Cannot create coefficient file %s\n")
		    call pargstr (coeffile)
		go to newspec_
		}
	    }

# Get next multispec number from list
	specap = -1
newap_
	if (nspecap <= 0) {
	    go to newspec_
	    }
	if (get_next_number (specap_range, specap) == EOF) {
	    go to newspec_
	    }
	if (cfd < 0) {
	    cfd = open (coeffile, APPEND, TEXT_FILE)
	    if (cfd < 2) {
		call eprintf ("SVDRES: Cannot append to coefficient file %s\n")
		    call pargstr (coeffile)
		go to newspec_
		}
	    }
	ispec = ispec + 1

	if (debug) {
	    call printf ("SVDRES: %s: ap= %d band= %d\n")
		call pargstr (specfile,SZ_PATHNAME)
		call pargi (specap)
		call pargi (specband)
	    }

	call eigenfit (specfile, specdir, specap, specband, nbin, fitmax,
		       rebspec, resspec, aptrack, fitspec, rebwl, reblines,
		       rebcont, ispec, nspec, imout, respath, cfd, verbose)

	if (oform == EIG_ONED || oform == EIG_MULTISPEC) {
	    call imunmap (imout)
	    imout = NULL
	    }

# Move on to next aperture or next image
	nspecap = nspecap - 1
	call close (cfd)
	cfd = -1
	if (nspecap > 0)
	    go to newap_
	go to newspec_
 
# Close the log files
endeig_	do i = 1, nlogfd {
	    call close (logfd[i])
	    }                                              

#  Close spectrum list
	call imtclose (speclist)
	if (imout != NULL && oform == EIG_TWOD)
	    call imunmap (imout)

#  Free memory used by spectrum buffers
	call mfree (rebspec, TY_REAL)
	call mfree (rebwl, TY_REAL)
	call mfree (rebcont, TY_REAL)
	call mfree (reblines, TY_REAL)
	call mfree (resspec, TY_REAL)
	call mfree (skyspec, TY_REAL)
	call mfree (fitspec, TY_REAL)
	if (debug) {
	    call printf ("SVDRES: object spectrum vectors freed\n")
	    call flush (STDOUT)
	    }
	if (conproc != NOCONT) {
	    call mfree (contvec, TY_REAL)
	    call mfree (contbuf, TY_REAL)
	    call mfree (contwts, TY_REAL)
	    call mfree (contrej, TY_REAL)
	    }
	if (debug) {
	    call printf ("SVDRES: object continuum vectors freed\n")
	    call flush (STDOUT)
	    }
	do ivec = 1, MAXTEMPS {
	    if (tmpsh[ivec] != NULL)
		call close_image (tmpim[ivec], tmpsh[ivec])
	    if (rebvec[ivec] != NULL)
		call mfree (rebvec[ivec], TY_REAL)
	    if (cnvec[ivec] != NULL)
		call mfree (cnvec[ivec], TY_REAL)
	    tmpsh[ivec] = NULL
	    tmpim[ivec] = NULL
	    rebvec[ivec] = NULL
	    cnvec[ivec] = NULL
	    }
	if (debug) {
	    call printf ("SVDRES: eigenvector vectors freed\n")
	    call flush (STDOUT)
	    }

end

# May  3 1999	New task, based on RVSAO t_xcsao.x
# Jul 13 1999	Make coefficient filename a parameter
# Sep  8 1999	Move to SVDFIT package
# Sep  9 1999	Make name for coefficient file
#
# Jan 18 2000	Pass continuum removal buffers in contin common
# Apr 28 3000	Move COUNTSPEC to separate file
# Jun 16 2000	Make residual directory if it does not already exist
# Oct 11 2000	Change parameter from output_format to res_format

# Jul 30 2003	Don't use nbin; it may not be set yet

# Dec 29 2005	Free rebvec and cnvec if set

# Jan  3 2005	Allocate and pass vector for object continuum
# Jan  4 2006	Do not write coefficient file if only one spectrum
# Jan  5 2006	Pass buffers as pointers so they can fit spectra if nbin=0
# Jan  5 2006	Add verbose parameter
# Jan  9 2006	Free eigenvector image, as well as WCS, structures
# Mar  3 2006	Set nbin to image first dimension if it is INDEF or zero
# Mar  6 2006	Fix debug statements
# Mar  7 2006	Free fitspec
# Jun 30 2006	Add eigentrack parameter so eigenvector aperture tracks object
# Nov 16 2006	Add fitmax parameter to set maximum flux ratio for sky fit

# May 29 2007	Add reblines vector to save chopped lines

# May 14 2008	Write out coefficients for each aperture as they are computed
# Jul  9 2008	Always set up coefficient file name

# Feb  5 2009	Set eigenvector continuum flag, which was never set before
