/*** File iolib/sbuffer.c
 *** August 20, 1998
 *** By Doug Mink
 *
 *  STPUT (SOURCE,OFFSET,NBYTES,DEST)
 *	Move bytes from numeric variable into character string
 *  STPUTC (SOURCE,OFFSET,NBYTES,DEST)
 *	Move bytes from a character string into a inter character string
 *  STGET (SOURCE,OFFSET,NBYTES,DEST)
 *	Move bytes from character string to variable (any numeric data type)
 *  STGETR8 (SOURCE OFFSET,DEST)
 *	Move bytes from character string to 8-byte IEEE floating point number
 *  STGETR4 (SOURCE OFFSET,DEST)
 *	Move bytes from character string to 4-byte IEEE floating point number
 *  STGETI4 (SOURCE OFFSET,DEST)
 *	Move bytes from character string to 4-byte integer
 *  STGETI2 (SOURCE OFFSET,DEST)
 *	Move bytes from character string to 2-byte integer
 *  STGETI1 (SOURCE,OFFSET,DEST)
 *	Move bytes from character string to Integer*1 variable
 *  STGETC (SOURCE,OFFSET,NBYTES,DEST)
 *	Move bytes from character string to another, smaller character string
 *  STGETL (SOURCE,OFFSET,NBYTES,DEST)
 *	Move one line from one character string to another
 *  STGETN (SOURCE,NLINE,NBYTES,DEST)
 *	Move one line from one character string to another
 *	If NLINE is =0, read next line
 */

#include <string.h>

static int iswap();

void
stput_ (source,offset,nbytes,dest,nc)

/*   Move bytes from numeric variable into character string
 *
 *	Call STPUT (SOURCE,OFFSET,NBYTES,DEST)
 *
 *	SOURCE	Address of variable from which to move first byte
 *
 *	OFFSET	Address of offset from start of string
 *
 *	NBYTES	Address of number of bytes to move
 *
 *	DEST	Address of character string to which to move first byte
 */
char *source,*dest;
int *offset,*nbytes;
int nc;
{
char *from, *last, *to;
	from = source;
	to = dest + *offset;
	last = from + *nbytes;
	if (nc < *offset) return;
	if (nc < *offset + *nbytes) last = from + nc;
	while (from < last) *(to++) = *(from++);
	return;
}


void
stputc_ (source,offset,nbytes,dest,ncs,ncd)

/*   Move bytes from a character string into a inter character string
 *
 *	Call STPUTC (SOURCE,OFFSET,NBYTES,DEST)
 *
 *	SOURCE	Address of character string from which to move first byte
 *
 *	OFFSET	Address of offset from start of string
 *
 *	NBYTES	Address of number of bytes to move
 *
 *	DEST	Address of character string to which to move bytes
 */
char *source,*dest;
int *offset,*nbytes;
int ncs,ncd;
{
char *from, *last, *to;
	from = source;
	to = dest + *offset;
	if (ncd < *offset) return;
	last = from + *nbytes;
	if (ncd < *offset + *nbytes) last = from + ncd;
	while (from < last) *(to++) = *(from++);
	return;
}


void
stget_ (source,offset,nbytes,dest,nc)

/*   Move bytes from character string to variable (any numeric data type)
 *
 *	Call STGET (SOURCE,OFFSET,NBYTES,DEST)
 *
 *	SOURCE	Address of character string from which to move first byte
 *
 *	OFFSET	Address of offset from start of string
 *
 *	NBYTES	Address of number of bytes to move
 *
 *	DEST	Address of variable to which to move first byte
 */
char *source,*dest;
int *offset,*nbytes;
int nc;
{
char *from, *last, *to;
	from = source + *offset;
	to = dest;
	last = from + *nbytes;
	if (nc < *offset) return;
	if (nc < *offset + *nbytes) last = from + nc;
	while (from < last) *(to++) = *(from++);
	return;
}


void
stgetc_ (source,offset,nbytes,dest,ncs,ncd)

/*   Move bytes from character string to another, smaller character string
 *
 *	Call STGETC (SOURCE,OFFSET,NBYTES,DEST)
 *
 *	SOURCE	Address of character string from which to move first byte
 *
 *	OFFSET	Address of offset from start of string
 *
 *	NBYTES	Address of number of bytes to move
 *
 *	DEST	Address of character string to which to move first byte
 */
char *source,*dest;
int *offset,*nbytes;
int ncs,ncd;
{
char *from, *last, *lastd, *to;

/*  Initialize pointers to start and end */
	from = source + *offset;
	to = dest;
	last = from + *nbytes;
	if (ncs < *offset) return;
	if (ncs < *offset + *nbytes) last = from + ncs;
	lastd = dest + (last - from);
	if (ncd < *lastd) last = from + ncd;

/*  Move bytes */
	while (from < last) *(to++) = *(from++);

/*  Pad destination with zeroes */
	if (*nbytes < ncd) {
	    lastd = dest + ncd;
	    while (to < lastd) *(to++) = 0;
	    }
	return;
}


void
stgetl_ (source,offset,nbytes,dest,ncs,ncd)

/*   Move one line from character string to another, smaller character string
 *
 *	Call STGETL (SOURCE,OFFSET,NBYTES,DEST)
 *
 *	SOURCE	Address of character string from which to move first byte
 *
 *	OFFSET	Address of offset from start of string (next line offset returned)
 *
 *	NBYTES	Address of number of bytes moved
 *
 *	DEST	Address of character string to which to move first byte
 */
char *source,*dest;
int *offset,*nbytes;
int ncs,ncd;
{
char *from, *last, *lastd, *to, lf;

/*  Initialize pointers to start and end */
	from = source + *offset;
	to = dest;
	last = from + ncd;
	if (ncs < *offset) return;
	if (ncs < *offset + ncd) last = from + ncs;
	lastd = dest + (last - from);
	if (ncd < *lastd) last = from + ncd;
	lf = 10;

/*  Move bytes */
	while (from < last && *from != lf) *(to++) = *(from++);
	*nbytes = to - dest;

/*  Pad destination with zeroes */
	if (*nbytes < ncd) {
	    lastd = dest + ncd;
	    while (to < lastd) *(to++) = 0;
	    }

/*  Set offset to next line */
	*offset = *offset + *nbytes + 1;
	return;
}

static char *next = NULL;


void
stgetn_ (source, nline, dest, nbytes, ncs, ncd)

/*   Move n'th line from character string to another, smaller character string
 *
 *	Call STGETN (SOURCE,NLINE,DEST,NBYTES)
 *
 *	SOURCE	Address of character string from which to move first byte
 *
 *	NLINE	Address of number of line (first = 1) to move
 *
 *	DEST	Address of character string to which to move first byte
 *
 *	NBYTES	Address of number of bytes moved
 */
char *source,*dest;
int *nline,*nbytes;
int ncs,ncd;
{
char	*from, *last, *lastd, *to, lf;
int	il, nl, ilf;

/*  Initialize pointers to start and end */
	from = source;
	to = dest;
	lf = 10;
	ilf = 10;
	nl = *nline;

	/* Move pointer to start of requested line */
	if (nl > 1) {
	    for (il = 0; il < nl; il ++) {
		next = strchr (from, ilf);
		if (next != NULL)
		    from = next + 1;
		else {
		    next = NULL;
		    *nbytes = 0;
		    return;
		    }
		}
	    }

	/* Set to first line if NLINE is 1 or next pointer is not set */
	else if (nl == 0 && next != NULL)
	    from = next;

	/* Use lengths of input and output buffers to set maximum line length */
	last = strchr (from, ilf);
	if (last == NULL) {
	    *nbytes = 0;
	    return;
	    }
	lastd = from + ncd;
	if (lastd < last) last = lastd;

	/*  Move bytes */
	while (from < last)
	    *(to++) = *(from++);

	/* Compute number of bytes in the returned string */
	*nbytes = to - dest;

	/* Set start of next line */
	if (*from == lf)
	    next = from + 1;

	/*  Pad destination with zeroes */
	if (*nbytes < ncd) {
	    *to = 0;
	    /* lastd = dest + ncd;
	    while (to < lastd) *(to++) = 0; */
	    }

	return;
}


void
stgetr8_ (source,offset,dest,nc)
char	*source;
double	*dest;
int	*offset;
int	nc;
{
char *from, *last, *to;
	from = source + *offset;
	if (iswap())		/* Swap bytes if necessary */
	    sswap8_ (from, 8);

	to =  (char *)dest;
	last = from + 8;
	while (from < last) *(to++) = *(from++);
	return;
}


void
stgetr4_ (source,offset,dest,nc)
char	*source;
float	*dest;
int	*offset;
int	nc;
{
char *from, *last, *to;
	from = source + *offset;
	if (iswap())		/* Swap bytes if necessary */
	    sswap4_ (from, 4);

	to =  (char *)dest;
	last = from + 4;
	while (from < last) *(to++) = *(from++);
	return;
}


void
stgeti4_ (source,offset,dest,nc)
char	*source;
int	*dest;
int	*offset;
int	nc;
{
char *from, *last, *to;
	from = source + *offset;
	if (iswap())		/* Swap bytes if necessary */
	    sswap4_ (from, 4);

	to =  (char *)dest;
	last = from + 4;
	while (from < last) *(to++) = *(from++);
	return;
}


void
stgeti2_ (source,offset,dest,nc)
char	*source;
short	*dest;
int	*offset;
int	nc;
{
char *from, *last, *to;
	from = source + *offset;
	if (iswap())		/* Swap bytes if necessary */
	    sswap2_ (from, 2);

	to =  (char *)dest;
	last = from + 2;
	while (from < last) *(to++) = *(from++);
	return;
}


stgeti1_ (source,offset,dest,nc,nc1)
char	*source;
char	*dest;
int	*offset;
int	nc;
int	nc1;
{
char *from, *last, *to;
	from = source + *offset;
	to =  (char *)dest;
	last = from + 1;
	*to = *from;
	return;
}


/* Check byte order */
static int
iswap()
{
short test = 1;
	if (*(char *)&test)
	    return (1);
	else
	    return (0);
}

/* Nov 25 1987	New subroutine
 *
 *Jan 15 1991	Combine STGET with STPUT into one source file
 *
 * Mar  3 1997	Add type-specific versions and byte swapping
 * Jun 10 1997	Add code to get a specific line
 *
 * Aug 20 1998	Add stget1()
 */
