/*** File iolib/stdio.c
 *** May 20, 1998
 *** By Doug Mink

 These subroutines allow Fortran 77 programs to use Unix standard I/O.
 There are 8 subroutines in this package:

	SOPEN:	 Opens a file for file I/O
	SCLOSE:	 Closes a file opened by SOPEN
	SLSEEK:	 Position file pointer
	SPOS:	 Position file pointer
	SGET:	 Reads from file
	SGETC:	 Reads character string from file
	SPUT:	 Writes to file
	SPUTC:	 Writes character string to file
	SPUTE:	 Writes character string to standard error
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

static int errno;

void
sopen_ (dev,flag,path,nc)

/*   Open UNIX file (Fortran-callable)
 *
 *	Call SOPEN (DEV,FLAG,PATH) from Fortran
 *
 *	Integer*4 DEV
 *			Address of UNIX file descriptor (returned)
 *	Integer*4 FLAG
 *			Address of read(=0)/write(=1)/create(=2)/update(=3) flag
 *	Character*(*) PATH
 *			Address of null-terminated file name (=path name)
 *
 *	nc	Number of characters in string PATH
 *		* this is not explicitly in the calling argument list;
 *		* Fortran puts it there, necessitating this variation on SPUT
 */

char *path;
int *flag;
int *dev;
int nc;

{
    int idev,flags,mode;
    extern int errno;

    /* Set flags for various kinds of file opening */
    if (*flag == 0) flags = O_RDONLY;
    if (*flag == 1) flags = O_WRONLY;
    if (*flag == 2) flags = O_RDWR + O_CREAT;
    if (*flag == 3) flags = O_RDWR;

    /* Set file open mode to write by user, read by everyone */
    mode = 0644;

    /* Open file */
    idev = open (path,flags,mode);
/*  printf ("openf: device is %d, flag is %d \n",idev,flags); */

    /* Set returned file descriptor */
    if (idev > 0) *dev = idev;
    else *dev = -errno;
	
    return;
}


void
sclose_ (dev,ier)

/*   Close UNIX file (Fortran-callable)
 *
 *	dev	Address of file descriptor
 *
 *	ier	Address of error code (returned, 0=OK)
 */
int *dev,*ier;
{
    int er,idev;
    static errno;

    idev = *dev;
    er = close (idev);
    if (er != 0) er = errno;
    if (er == 0) *dev = -2;
    *ier = er;
    return;
}


void
slseek_ (dev,offset,whence,pos)

/*   Position UNIX file pointer (Fortran-callable)
 *   (See LSEEK(2) in UNIX Programmer's Manual)
 *
 *	dev	Address of UNIX file descriptor
 *		Integer*4
 *
 *	offset	Address of number of bytes to offset
 *		Integer*4
 *
 *	whence	Address of pointer extension type
 *		0 > set to offset, 1 > current+offset, 2 > end+offset
 *		Integer*4
 *
 *	pos	Address of file pointer (returned; should = offset)
 *		Integer*4
 */
int *dev,*offset,*whence,*pos;
{
    int idev;

    idev = *dev;
    *pos = lseek (idev,*offset,*whence);
    return;
}


void
spos_ (dev,offset,pos)

/*   Position UNIX file pointer (Fortran-callable)
 *
 *	dev	Address of UNIX file descriptor
 *
 *	offset	Address of number of bytes to offset
 *
 *	pos	Address of file pointer (returned; should = offset)
 */
int *dev,*offset,*pos;
{
    int whence,idev;

    whence = 0;
    idev = *dev;
    *pos = lseek (idev,*offset,whence);
    return;
}


void
sget_ (dev,nbytes,buf,nread)

/*   Read from UNIX file (Fortran-callable)
 *
 *	dev	Address of UNIX file descriptor
 *
 *	nbytes	Address of number of bytes to read
 *
 *	buf	Address of buffer array (values returned)
 *
 *	nread	Address of number of bytes read (value returned)
 */
char *buf;
int *nbytes,*dev,*nread;
{
    int idev,nb,nr;
    extern int errno;

    idev = *dev;
    nb = *nbytes;
    nr = read (idev,buf,nb);

    if (nr > -1) *nread = nr;
    else *nread = -errno;

    return;
}


void
sgetc_ (dev,nbytes,buf,nread,nc)

/*   Read character string from UNIX file (Fortran-callable)
 *
 *	dev	Address of UNIX file descriptor
 *		Integer*4
 *
 *	nbytes	Address of number of bytes to read
 *		Integer*4
 *
 *	buf	Address of buffer array (values returned)
 *		Character*(*)
 *
 *	nread	Address of number of bytes read (value returned)
 *		Integer*4
 *
 *	nc	Number of characters in string buf
 *		* This is not explicitly in the calling argument list;
 *		* Fortran puts it there, necessitating this variation on SGET
 */
char *buf;
int *nbytes,*dev,*nread,nc;
{
    int idev,nb,nr;
    extern int errno;

    idev = *dev;
    nb = *nbytes;
    nr = read (idev,buf,nb);

    if (nr > -1) *nread = nr;
    else *nread = -errno;

    return;
}


void
sput_ (dev,nbytes,buf,nwrit)

/*   Write to UNIX file (Fortran-callable)
 *
 *	dev	Address of UNIX file descriptor
 *
 *	nbytes	Address of number of bytes to read
 *
 *	buf	Address of buffer array (values returned)
 *
 *	nwrit	Address of number of bytes written (returned, should = nbytes)
 */
char *buf;
int *nbytes,*dev,*nwrit;
{
    int nw,idev,nb;
    extern int errno;

    idev = *dev;
    nb = *nbytes;
    nw = write (idev,buf,nb);

    if (nw > -1) *nwrit = nw;
    else *nwrit = -errno;

    return;
}


void
sputc_ (dev,nbytes,buf,nwrit,nc)

/*   Write character string to UNIX file (Fortran-callable)
 *
 *	dev	Address of UNIX file descriptor
 *		Integer*4
 *
 *	nbytes	Address of number of bytes to read
 *		Integer*4
 *
 *	buf	Address of buffer array (values returned)
 *		Character*(*)
 *
 *	nwrit	Address of number of bytes written (returned, should = nbytes)
 *		Integer*4
 *
 *	nc	Number of characters in string buf
 *		* This is not explicitly in the calling argument list;
 *		* Fortran puts it there, necessitating this variation on SPUT
 */
char *buf;
int *nbytes,*dev,*nwrit,nc;
{
    int nw,idev,nb;
    extern int errno;

    idev = *dev;
    nb = *nbytes;
    nw = write (idev,buf,nb);

    if (nw > -1) *nwrit = nw;
    else *nwrit = -errno;

    return;
}


void
spute_ (nbytes,buf,nwrit,nc)

/*   Write character string to UNIX standard error (Fortran-callable)
 *
 *	nbytes	Address of number of bytes to read
 *		Integer*4
 *
 *	buf	Address of buffer array (values returned)
 *		Character*(*)
 *
 *	nwrit	Address of number of bytes written (returned, should = nbytes)
 *		Integer*4
 *
 *	nc	Number of characters in string buf
 *		* This is not explicitly in the calling argument list;
 *		* Fortran puts it there, necessitating this variation on SPUT
 */
char *buf;
int *nbytes,*nwrit,nc;
{
    int nw,idev,nb;
    extern int errno;

    nb = *nbytes;
    nw = write (2,buf,nb);

    if (nw > -1) *nwrit = nw;
    else *nwrit = -errno;

    return;
}

/* Aug 27 1993	Change longs to ints
 * 
 * May 20 1998	Declare all subroutines void
 */
