/*** File iolib/stream.c
 *** November 8, 2001
 *** By Dick French, modified by Doug Mink

 These subroutines allow Fortran 77 programs to use buffered stream I/O.
 There are 10 subroutines in this package:

	SFOPEN:	 Opens a file for stream I/O
	SFCLOSE: Closes a file opened by SFOPEN
	SFSEEK:	 Position file pointer in stream file
	SFPOS:	 Position file pointer in stream file
	SFTELL:	 Return file pointer position
	SFLUSH:  Flush buffer
	SFGET:	 Reads from stream
	SFGETC:	 Reads Character string from stream
	SFPUT:	 Writes to stream
	SFPUTC:	 Writes Character string to stream
	SFGETL:	 Reads Character string from one line of stream
	SFPUTL:	 Writes Character string followed by a NEWLINE to stream
 */

#include <sys/file.h>
#include <stdio.h>
#include <string.h>

FILE *streams[10];
int nextstream;

void
sfopen_ (idev,iflag,path,nc)

/*   Open UNIX file (Fortran-callable) for buffered I/O using system routines.
 *
 *	Call SFOPEN (IDEV,IFLAG,PATH)
 *
 *	Integer*4 IDEV
 *		Address of UNIX file descriptor (returned) (-1 if open failed)
 *	Integer*4 IFLAG
 *		Address of iflag :
 *		0	open for reading
 *		1	create for writing
 *		2	create for reading and writing at start of file
 *		3	append: open for writing at end of file, or create
 *			for writing
 *	Character*(*) PATH
 *			Address of null-terminated file name (=path name)
 *
 *	nc	Number of characters in string PATH
 *		* this is not explicitly in the calling argument list;
 *		* Fortran puts it there, necessitating this variation on SPUT
 */

char *path;
int *iflag;
int *idev;
int nc;

{
    char *type;
    FILE *stream;
    int id;

    if (*iflag == 0) type = "r";
    if (*iflag == 1) type = "w";
    if (*iflag == 2) type = "w+";
    if (*iflag == 3) type = "a";

    stream = fopen (path,type);

    if (stream == NULL)
	*idev = -1; /* set device to -1 if open failed*/
    else {
	for (id = 0; id < 10; id++) {
	    if (streams[id] == NULL) {
		*idev = id+3;
		streams[id] = stream;
		break;
		}
	    }
	}

    return;
}


void
sflush_ (idev)

/*   Flush buffer to UNIX file opened for buffered I/O (Fortran-callable)
 *
 *   Fortran calling sequence :  Call SFLUSH (IDEV)
 *
 *	IDEV	Address of UNIX file descriptor (input)
 */

int *idev;

{
    FILE *stream;
    int ptr;

    stream = streams[*idev-3];
    ptr=fflush (stream);
	
    return;
}


void
sfclose_ (idev,ier)

/*   Close UNIX file opened for buffered I/O (Fortran-callable)
 *
 *   Fortran calling sequence :  Call SFCLOSE (IDEV,IER)
 *
 *	IDEV	Address of UNIX file descriptor (input)
 *
 *	IER	Address of returned error flag (0 => OK)
 */

int *idev;
int *ier;

{
    FILE *stream;
    int ptr;

    stream = streams[*idev-3];
    ptr=fclose(stream);
    streams[*idev] = NULL;
    *ier=ptr;

    return;
}


void
sfseek_ (idev,noff,nptr,ier)

/*   Reposition a UNIX file (Fortran-callable) using FSEEK(3S)
 *
 *   Call SFSEEK (IDEV,NOFF,NPTR,IER)
 *
 *	IDEV	Address of UNIX file descriptor
 *
 *	NOFF	Address of number of bytes to offset
 *
 *	NPTR	Address of nptr. The new position is at the signed distance
 *		noff bytes from the beginning, the current position, or the
 *		end of the file, according as nptr has the value 0,1,or 2.
 *
 *	IER	Address of returned error flag (-1 for improper seek)
 */

int *idev,*noff,*nptr,*ier;

{
    FILE *stream;
    int offset,ptrname;
    int nier;

    stream = streams[*idev-3];
    offset = *noff;
    ptrname = *nptr;
    nier = fseek(stream,offset,ptrname);
    *ier = nier;	
    return;
}

void
sfpos_ (idev, noff, pos)

/*   Reposition a UNIX file (Fortran-callable) using FSEEK(3S) from start
 *
 *   Call SFPOS (IDEV,NOFF,POS)
 *
 *	IDEV	Address of UNIX file descriptor
 *
 *	NOFF	Address of number of bytes to offset
 *
 *	POS	Address of file pointer (returned; should = offset)
 */

int *idev,*noff,*pos;

{
    FILE *stream;
    int offset;
    int nier;

    stream = streams[*idev-3];
    offset = *noff;
    nier = fseek (stream, offset, SEEK_SET);
    *pos = ftell(stream);
    return;
}


void
sftell_ (idev,noff)

/*   Return current position of a UNIX file (Fortran-callable) using ftell
 *
 *   Call SFTELL (IDEV,NOFF)
 *
 *	IDEV	Address of UNIX file descriptor (input)
 *
 *	NOFF	Address of current value of the offset relative to the 
 *		beginning of the file, in bytes (returned)
 */

int *idev,*noff;

{
    FILE *stream;
    int nnoff;

    stream = streams[*idev-3];
    nnoff = ftell(stream);
    *noff = nnoff;

    return;
}


void
sfget_ (idev,nbytes,buf,nread)

/*   Read from UNIX file (Fortran-callable) using buffered I/O.
 *
 *   Call SFGET (IDEV,NBYTES,BUF,NREAD)
 *
 *	IDEV	Address of UNIX file descriptor (input)
 *
 *	NBYTES	Address of number of bytes to read (input)
 *
 *	BUF	Address of buffer array (values returned)
 *
 *	NREAD	Address of number of bytes read (returned, should=nbytes)
 */

char *buf;
int *nbytes,*nread;
int *idev;

{
    FILE *stream;
    int nr,nitems,size;

    nitems = *nbytes ;
    size = 1;
    stream = streams[*idev-3];
    nr = fread(buf,size,nitems,stream);
    *nread = nr;

    return;
}


void
sfgetc_ (idev,nbytes,buf,nread,nc)

/*   Read character string from UNIX file (Fortran-callable) using buffered I/O
 *
 *   Call SFGET (IDEV,NBYTES,BUF,NREAD)
 *
 *	IDEV	Address of UNIX file descriptor (input)
 *		Integer*4
 *
 *	NBYTES	Address of number of bytes to read (input)
 *		Integer*4
 *
 *	BUF	Address of buffer array (values returned)
 *		Character*(*)
 *
 *	NREAD	Address of number of bytes read (returned, should=nbytes)
 *		Integer*4
 *
 *	nc	Number of characters in string buf
 *		* This is not explicitly in the calling argument list;
 *		* Fortran puts it there, necessitating this variation on SPUT
 */

char *buf;
int *nbytes,*nread,*idev,nc;

{
    FILE *stream;
    int nr,nitems,size;

    nitems = *nbytes ;
    size = 1;
    stream = streams[*idev-3];
    nr = fread(buf,size,nitems,stream);
    *nread = nr;
	
    return;
}


void
sfgetl_ (idev,buf,nread,nc)

/*   Read one line from UNIX file (Fortran-callable)
 *
 *   Call SFGETL (IDEV,BUF,NREAD)
 *
 *	IDEV	Address of UNIX file descriptor (input)
 *		Integer*4
 *
 *	BUF	Address of buffer array (values returned)
 *		Character*(*)
 *
 *	NREAD	Address of number of bytes read (returned)
 *		Integer*4
 *
 *	nc	Number of characters in string buf
 *		* This is not explicitly in the calling argument list;
 *		* Fortran puts it there all by itself.
 */
char *buf;
int *idev, *nread, nc;
{
    FILE *stream;
    char nl, cr, *last, *change, *errc;
    int nr, nch, i;

    stream = streams[*idev-3];
    nl = 10;
    cr = 13;
    nr = nc;
    for (i = 0; i < nr; i++) buf[i] = 0;
    errc = fgets (buf,nr,stream);
    /* if (errc == NULL) {
	*nread = -1;
	return;
	} */
    /* printf ("SFGETL: string = %s\n",buf); */
    last = strchr (buf,nl);
    if (last != NULL) {
	nch = last - buf;
	*last = 0;
	}
    else if (*buf == 0) {
	nch = 0;
	for (i = 0; i < nr; i++) buf[i] = 0;
	}
    else {
	nch = nr;
	}
    if (*buf == cr && nch > 0) {
	last = buf + nch - 1;
	change = buf;
	while (change < last) {
	    *(change+1) = *change;
	    change++;
	    }
	}
    *nread = nch;

    return;
}


void
sfput_ (idev,nbytes,buf,nwrit)

/*   Write to UNIX file (Fortran-callable) using buffered I/O.  
 *
 *   Call SFPUT (IDEV,NBYTES,BUF,NWRIT)
 *
 *	IDEV	Address of UNIX file descriptor (input)
 *
 *	NBYTES	Address of number of bytes to write (input)
 *
 *	BUF	Address of buffer array (input values to write to file)
 *
 *	NWRIT	Address of number of bytes written (returned, should=nbytes)
 */

char *buf;
int *nbytes,*nwrit;
int *idev;

{
    FILE *stream;
    int nw,nitems,size;

    nitems = *nbytes;
    size = 1;
    stream = streams[*idev-3];
    nw = fwrite(buf,size,nitems,stream);
    *nwrit = nw;

    return;
}


void
sfputc_ (idev,nbytes,buf,nwrit,nc)

/*   Write character string to UNIX file (Fortran-callable) using buffered I/O.  
 *
 *   Call SFPUT (IDEV,NBYTES,BUF,NWRIT)
 *
 *	IDEV	Address of UNIX file descriptor (input)
 *		Integer*4
 *
 *	NBYTES	Address of number of bytes to write (input)
 *		Integer*4
 *
 *	BUF	Address of buffer array (input values to write to file)
 *		Character*(*)
 *
 *	NWRIT	Address of number of bytes written (returned, should=nbytes)
 *		Integer*4
 *
 *	nc	Number of characters in string buf
 *		* This is not explicitly in the calling argument list;
 *		* Fortran puts it there, necessitating this variation on SPUT
 */

char *buf;
int *nbytes,*nwrit,*idev,nc;

{
    FILE *stream;
    int nw,nitems,size;

    nitems = *nbytes ;
    size = 1;
    stream = streams[*idev-3];
    nw = fwrite(buf,size,nitems,stream);
    *nwrit = nw;
	
    return;
}


void
sfputl_ (idev,buf,nc)

/*   Write one line to UNIX file (Fortran-callable)
 *
 *   Call SFPUTL (IDEV,BUF)
 *
 *	IDEV	Address of UNIX file descriptor (input)
 *		Integer*4
 *
 *	BUF	Address of buffer array (values returned)
 *		Character*(*)
 *
 *	nc	Number of characters in string buf
 *		* This is not explicitly in the calling argument list;
 *		* Fortran puts it there all by itself.
 */
char *buf;
int *idev,nc;
{
    FILE *stream;
    char c[2];

    /* Write null-terminated string */
    stream = streams[*idev-3];
    fputs (buf,stream);

    /* Add newline */
    c[0] = 10;
    c[1] = 0;
    fputs (c,stream);

    return;
}

/* Nov  7 1988	Add fflush(stream) as sflush subroutine
 * Oct 11 1990	Fix sfgetl to ignore CR after LF
 * May  4 1993	Call strchr instead of index for System 5 compatibility
 * Aug 27 1993	Change longs to ints
 * Sep 15 1993	Keep track of stream pointers internally
 *
 * Oct 17 1995	Return NREAD=-1 if EOF reached by SFGETL
 *
 * May 20 1998	Declare all subroutines void
 *
 * Feb 27 2001	Add sfpos() to position relative to start of file
 * Nov  8 2001	Fix ambiguous statement in sfgetl()
*/
